'use strict';

import elt from '../../../../../node_modules/crel/crel';

const prefix = "ProseMirror-prompt";

export function openPrompt(view, options) {
  let wrapper = document.createElement("div");
  wrapper.className = prefix;
  let hadFocus = view.hasFocus();

  let mouseOutside = e => {
    if (!wrapper.contains(e.target)) {
      close();
    }
  };
  setTimeout(() => window.addEventListener("mousedown", mouseOutside), 50);
  let close = () => {
    window.removeEventListener("mousedown", mouseOutside);
    if (wrapper.parentNode) wrapper.parentNode.removeChild(wrapper);
    if (hadFocus) setTimeout(() => view.focus(), 50);
  };

  let domFields = [];
  for (let name in options.fields) domFields.push(options.fields[name].render());

  let submitButton = document.createElement("button");
  submitButton.type = "submit";
  submitButton.className = prefix + "-submit";
  submitButton.textContent = "OK";
  let cancelButton = document.createElement("button");
  cancelButton.type = "button";
  cancelButton.className = prefix + "-cancel";
  cancelButton.textContent = "Cancel";
  cancelButton.addEventListener("click", close);

  let form = wrapper.appendChild(document.createElement("form"));
  if (options.title) form.appendChild(document.createElement("h5")).textContent = options.title;
  domFields.forEach(field => {
    form.appendChild(document.createElement("div")).appendChild(field);
  });
  let buttons = form.appendChild(document.createElement("div"));
  buttons.className = prefix + "-buttons";
  buttons.appendChild(submitButton);
  buttons.appendChild(document.createTextNode(" "));
  buttons.appendChild(cancelButton);

  let outerBox = view.dom.parentNode.getBoundingClientRect();
  view.dom.parentNode.appendChild(wrapper);
  let blockBox = wrapper.getBoundingClientRect();
  let cX = Math.max(0, outerBox.left) + Math.min(window.innerWidth, outerBox.right) - blockBox.width;
  let cY = Math.max(0, outerBox.top) + Math.min(window.innerHeight, outerBox.bottom) - blockBox.height;
  wrapper.style.left = (cX / 2 - outerBox.left) + "px";
  wrapper.style.top = (cY / 2 - outerBox.top) + "px";

  let submit = () => {
    let params = getValues(options.fields, domFields);
    if (params) {
      close();
      options.callback(params);
    }
  };

  form.addEventListener("submit", e => {
    e.preventDefault();
    submit();
  });

  form.addEventListener("keydown", e => {
    if (e.keyCode == 27) {
      e.preventDefault();
      close();
    } else if (e.keyCode == 13 && !(e.ctrlKey || e.metaKey || e.shiftKey)) {
      e.preventDefault();
      submit();
    } else if (e.keyCode == 9) {
      window.setTimeout(() => {
        if (!wrapper.contains(document.activeElement)) close();
      }, 500);
    }
  });

  let input = form.elements[0];
  if (input) input.focus();
}

function getValues(fields, domFields) {
  let result = Object.create(null), i = 0;
  for (let name in fields) {
    let field = fields[name], dom = domFields[i++];
    let value = field.read(dom), bad = field.validate(value);
    if (bad) {
      reportInvalid(dom, bad);
      return null;
    }
    result[name] = field.clean(value);
  }
  return result;
}

function reportInvalid(dom, message) {
  // FIXME this is awful and needs a lot more work
  let parent = dom.parentNode;
  let msg = parent.appendChild(document.createElement("div"));
  msg.style.left = (dom.offsetLeft + dom.offsetWidth + 2) + "px";
  msg.style.top = (dom.offsetTop - 5) + "px";
  msg.className = "ProseMirror-invalid";
  msg.textContent = message;
  setTimeout(() => parent.removeChild(msg), 1500);
}

// ::- The type of field that `FieldPrompt` expects to be passed to it.
export class Field {
  // :: (Object)
  // Create a field with the given options. Options support by all
  // field types are:
  //
  // **`value`**`: ?any`
  //   : The starting value for the field.
  //
  // **`label`**`: string`
  //   : The label for the field.
  //
  // **`required`**`: ?bool`
  //   : Whether the field is required.
  //
  // **`validate`**`: ?(any) → ?string`
  //   : A function to validate the given value. Should return an
  //     error message if it is not valid.
  constructor(options) { this.options = options; }

  // render:: (state: EditorState, props: Object) → dom.Node
  // Render the field to the DOM. Should be implemented by all subclasses.

  // :: (dom.Node) → any
  // Read the field's value from its DOM node.
  read(dom) { return dom.value; }

  // :: (any) → ?string
  // A field-type-specific validation function.
  validateType(_value) {}

  validate(value) {
    if (!value && this.options.required)
      return "Required field";
    return this.validateType(value) || (this.options.validate && this.options.validate(value));
  }

  clean(value) {
    return this.options.clean ? this.options.clean(value) : value;
  }
}

// ::- A field class for single-line text fields.
export class TextField extends Field {
  render() {
    let input = document.createElement("input");
    input.type = "text";
    input.placeholder = this.options.label;
    input.value = this.options.value || "";
    input.autocomplete = "off";
    return input;
  }
}


// ::- A field class for dropdown fields based on a plain `<select>`
// tag. Expects an option `options`, which should be an array of
// `{value: string, label: string}` objects, or a function taking a
// `ProseMirror` instance and returning such an array.
export class SelectField extends Field {
  render() {
    let select = document.createElement("select");
    this.options.options.forEach(o => {
      let opt = select.appendChild(document.createElement("option"));
      opt.value = o.value;
      opt.selected = o.value == this.options.value;
      opt.label = o.label;
    });
    return select;
  }
}



/**********
 * WIP (Do not use)
 *********/

// Build a form based on field (Fields)
export class MenuForm {
  constructor(view, title, fields) {
    this.view = view;
    this.title = title;
    this.fields = fields;
    this.domFields = [];
    for (let name in fields) {
      this.domFields.push(fields[name].render(view));
    }

    let titleEl = elt("h5", {}, view.translate(title));
    let submitButton = elt("button", {type: "submit", class: "ProseMirror-form-submit"}, view.translate("Add"));
    let cancelButton = elt("button", {type: "button", class: "ProseMirror-form-cancel"}, view.translate("Cancel"));
    cancelButton.addEventListener("click", () => view.props.on.interaction.dispatch());


    this.form = elt("form", null, titleEl, this.domFields.map(f => elt("div", null, f)),
                    elt("div", {class: "ProseMirror-form-buttons"}, submitButton, " ", cancelButton));
  }

  render(view, callback) {
    let submit = () => {
      let params = this.values();
      if (params) {
        callback(params);
      }
    };

    this.form.addEventListener("submit", e => {
      e.preventDefault();
      e.stopPropagation();
      submit();
    });

    this.form.addEventListener("keydown", e => {
      if (e.keyCode == 27) {
        e.preventDefault();
        view.props.on.interaction.dispatch();
      } else if (e.keyCode == 13 && !(e.ctrlKey || e.metaKey || e.shiftKey)) {
        e.preventDefault();
        submit();
      }
    });

    let input = this.form.elements[0];
    if (input) input.focus();

    return this.form;
  }

  values() {
    let result = Object.create(null), i = 0;
    for (let name in this.fields) {
      let field = this.fields[name], dom = this.domFields[i++];
      let value = field.read(dom), bad = field.validate(value);
      if (bad) {
        this.reportInvalid(dom, this.view.translate(bad));
        return null;
      }
      result[name] = field.clean(value);
    }
    return result;
  }

  reportInvalid(dom, message) {
    let parent = dom.parentNode;
    let style = "left: " + (dom.offsetLeft + dom.offsetWidth + 2) + "px; top: " + (dom.offsetTop - 5) + "px";
    let msg = parent.appendChild(elt("div", {class: "ProseMirror-invalid", style}, message));
    setTimeout(() => parent.removeChild(msg), 1500);
  }
}