'use strict';

function StoreData(indexProperties, autoIdPrefix) {
    this.ids = {};
    this.indexProperties = indexProperties || [];
    this.autoIdPrefix = autoIdPrefix || 'auto-';
    this.indexes = {};
    this.missingIdCounter = 0;
};

StoreData.prototype = new Array();

StoreData.prototype.exists = function(id) {
  return typeof(this.ids[id]) !== 'undefined';
};

StoreData.prototype.count = function() {
  return this.length;
};

StoreData.prototype.all = function() {
  return this.slice(0);
};

StoreData.prototype.get = function(id) {
  return this[this.ids[id]];
};

StoreData.prototype.push = function(obj) {
  if (typeof(obj.id) === 'undefined') {
    var autoId;
    while (!autoId || typeof(this.ids[autoId]) !== 'undefined') {
      autoId = this.autoIdPrefix + this.missingIdCounter++;
    }
    obj.id = autoId;
  }
  if (typeof(this.ids[obj.id]) !== 'undefined') {
    throw new StoreError('Duplicate id');
  }
  this.ids[obj.id] = this.length;
  Array.prototype.push.call(this, obj);
  this._index(obj);
};

StoreData.prototype.update = function(obj, id) {
  if (id === undefined) {
    id = obj.id;
  }
  obj = angular.extend(this[this.ids[id]], obj);
  if (id !== obj.id) {
    this.ids[obj.id] = this.ids[id];
    delete this.ids[id];
  }
  this._index(obj);
  return obj;
};

StoreData.prototype.remove = function(obj) {
  if (!this.exists(obj.id)) {
    throw new StoreNotFoundError('Object does not exist');
  }
  var id = obj.id;
  var idx = this.ids[id];
  obj = this[idx];
  this._deindex(obj);
  this.splice(idx, 1);
  delete this.ids[id];
  this.rearrangeIdsAfterRemove(idx);
  return obj;
};

StoreData.prototype.rearrangeIdsAfterRemove = function(removedIdx) {
  angular.forEach(this.ids, function(idx, id) {
    if (idx > removedIdx) {
      this.ids[id] = idx - 1;
    }
  }, this);
};

StoreData.prototype.replace = function(objects) {
  this.splice(0, this.length);
  this.ids = {};
  this.indexes = {};
  this.missingIdCounter = 0;
  objects.forEach(this.push, this);
  this.missingIdCounter = this.length;
};

StoreData.prototype.ids = function() {
  return Object.keys(this.ids);
};

StoreData.prototype.getIndexes = function() {
  return this.indexes;
};

StoreData.prototype.getIndexDimensions = function(key) {
  if (this.indexProperties.indexOf(key) === -1) {
    throw new StoreNotFoundError('Missing index for key "' + key + '"');
  }
  if (typeof(this.indexes[key]) === 'undefined') {
    return [];
  }
  return Object.keys(this.indexes[key]);
};

StoreData.prototype._index = function(obj) {
  this._iterIndexProperties(function(key, index) {
    var found = false;
    this._clearObjectInIndex(obj, index);
    if (typeof(obj[key]) !== 'undefined') {
      if (typeof(index[obj[key]]) === 'undefined') {
        index[obj[key]] = [];
      }
      index[obj[key]].push(obj);
    }
  });
};

StoreData.prototype._deindex = function(obj) {
  this._iterIndexProperties(function(key, index) {
    this._clearObjectInIndex(obj, index);
  });
};

StoreData.prototype._clearObjectInIndex = function(obj, index) {
  angular.forEach(index, function(objects, value) {
    var idx = objects.indexOf(obj);
    if (idx > -1) {
      objects.splice(idx, 1);
    }
  });
};

StoreData.prototype._iterIndexProperties = function(callback) {
  this.indexProperties.forEach(function(key) {
    if (typeof(this.indexes[key]) === 'undefined') {
      this.indexes[key] = {};
    }
    callback.apply(this, [key, this.indexes[key]]);
  }, this);
};
