'use strict';

var StoreLoader = function(store, baseServiceParams) {
  this.store = store;
  this.called = [];
  if (baseServiceParams) {
    this.baseServiceParams = baseServiceParams;
  }
};

StoreLoader.prototype.baseServiceParams = null;

StoreLoader.prototype.remoteServiceDataProperty = null;

StoreLoader.prototype._prepareParams = function(params) {
  if (this.baseServiceParams) {
    return angular.extend({}, params, this.baseServiceParams);
  }
  return params || {};
};

StoreLoader.prototype._getRequestSignature = function(action, params) {
  return action.__name__ + JSON.stringify(params);
};

StoreLoader.prototype._lock = function(action, params, callback, loader) {
  return this.store._lock(this._getRequestSignature(action, params), callback, loader);
};

StoreLoader.prototype._request = function(action, params, callback, errorCallback) {
  var self = this;

  if (typeof(params) === 'function') {
    // possibility to omit params in arguments
    errorCallback = callback;
    callback = params;
    params = {};
  }

  params = this._prepareParams(params);
  this._lock(action, params, callback, function(success, error, lock) {
    action.prepare(params).get(success, function(r) {
      var lockCallbacksCount = lock.callbacks.length;
      error();
      if (errorCallback) {
        while (lockCallbacksCount > 1) {
          errorCallback(r);
          lockCallbacksCount -= 1;
        }
        return errorCallback(r);
      }
      return self.$q.reject();
    }, {
      ignoreLoadingBar: false
    });
  });
};

StoreLoader.prototype._processObject = function(obj) {
  return obj;
};

StoreLoader.prototype._loadMany = function(action, params, callback, errorCallback) {
  if (typeof(params) === 'function') {
    errorCallback = callback;
    callback = params;
    params = {};
  }

  var self = this;
  this._request(action, params, function(objects, response) {
    self._processLoadManyResponse(objects, self.remoteServiceDataProperty, function(objects) {
      self.called[self._getRequestSignature(action, params)] = objects;
      callback && callback(objects, response);
    });
  }, errorCallback);
};

StoreLoader.prototype._processLoadManyResponse = function(objects, dataProperty, callback) {
  if (dataProperty) {
    objects = objects[dataProperty];
  }
  var self = this;
  return this.store.transaction(function() {
    objects = objects.map(function(obj) {
      return this.upsert(self._processObject(obj));
    }, this);
    callback && callback(objects);
  });
};

StoreLoader.prototype._loadManyOnce = function(action, params, callback, errorCallback) {
  var sig = this._getRequestSignature(action, params)
  if (typeof(this.called[sig]) !== 'undefined') {
    callback && callback(this.called[sig]);
  } else {
    this._loadMany(action, params, callback, errorCallback);
  }
};

StoreLoader.prototype._load = function(action, params, callback, errorCallback) {
  if (typeof(params) === 'function') {
    errorCallback = callback;
    callback = params;
    params = {};
  }
  var self = this;
  this._request(action, params, function(obj, response) {
    self._processLoadResponse(obj, function(obj) {
      callback && callback(obj, response);
    });
  }, errorCallback);
};

StoreLoader.prototype._processLoadResponse = function(obj, callback) {
  obj = this.store.upsert(this._processObject(obj));
  callback && callback(obj);
};

StoreLoader.prototype._getOrLoad = function(id, action, params, callback, errorCallback) {
  var obj;
  try {
    obj = this.store.get(id);
  } catch (e) {
    if (!(e instanceof StoreNotFoundError)) {
      throw e;
    }
    this._load(action, params, callback, errorCallback);
    return;
  }
  if (typeof(params) === 'function') {
    params(obj);
  } else if (callback) {
    callback(obj);
  }
};

StoreLoader.prototype._getter = function(func, getter, loader) {
  var self = this;
  getter = getter || this._getOrLoad;
  if (!loader) {
    loader = function() {
      var args = Array.from(arguments).slice(1);
      return self._load.apply(self, args);
    };
  }

  var wrapper = function() {
    var args = Array.from(arguments);
    return func.apply(self, [getter.bind(self)].concat(args));
  };
  wrapper.load = function() {
    var args = Array.from(arguments);
    return func.apply(self, [loader.bind(self)].concat(args));
  };
  return wrapper;
};

StoreLoader.prototype._getterMany = function(func) {
  return this._getter(func, this._loadManyOnce, this._loadMany);
};

StoreLoader.async = function(arr, callback, errorCallback, thisArg) {
  if (arr.length === 0) {
    callback && callback([]);
    return;
  }
  var results = [];
  var lock = arr.length - 1;
  var failed = false;
  var cb = function(result) {
    if (result !== undefined) {
      results.push(result);
    }
    if (--lock < 0 && !failed) {
      callback && callback(results);
    }
  };
  var errcb = function() {
    if (!failed) {
      errorCallback && errorCallback.apply(null, arguments);
      failed = true;
    }
  };
  arr.forEach(function(func) {
    if (typeof(func) === 'function') {
      func.call(thisArg || null, cb, errcb);
    } else {
      var args = func.slice(1).concat([cb, errcb]);
      func[0].apply(thisArg || null, args);
    }
  });
};

StoreLoader.asyncForEach = function(arr, iterator, callback, errorCallback, thisArg) {
  var calls = arr.map(function(item) { return [iterator, item]});
  return StoreLoader.async(calls, callback, errorCallback, thisArg);
};
