'use strict';

var StoreQuery = function(data, filters, processor, indexes) {
  this.data = data;
  this.filters = filters;
  if (filters instanceof StoreQuery) {
    this.filters = filters.filters;
  }
  this.processor = processor;
  this.indexes = indexes;
  this.optimized = this._optimize();
};

StoreQuery.prototype.filter = function(filters) {
  this.filters = StoreQuery.mergeFilters(this.filters, filters);
  return this;
};

StoreQuery.prototype.execute = function() {
  var matches, metas;

  if (!this.filters) {
    return this._results(this.data.slice(0));
  }

  if (typeof(this.filters) === 'array') {
    return this._results(this._intersection.apply(this, this.filters.map(function(filters) {
      return this.query(filters);
    }, this)));
  }

  if (typeof(this.filters) === 'function') {
    matches = this.data.filter(this.filters);
  } else if (this.optimized.indexes) {
    matches = this._filterByIndexes(this.optimized.indexes);
  } else {
    matches = this.data; // no need for shallow copy as we filter the array
  }

  if (this.optimized.filters) {
    metas = {};
    matches = matches.filter(function(obj) {
      var meta = {};
      if (StoreQuery.matchObject(obj, this.optimized.filters, meta)) {
        metas[obj.id] = meta;
        return true;
      }
      return false;
    }, this);
  } else {
    matches = matches.slice(0);
  }

  return this._results(matches, metas);
};

StoreQuery.prototype._results = function(matches, meta) {
  if (this.processor) {
    matches = matches.map(this.processor);
  }
  return new StoreResults(matches, meta, this);
};

StoreQuery.prototype._optimize = function() {
  var spec = {indexes: null, filters: null};

  if (!this.filters || typeof(this.filters) === 'function') {
    return spec;
  }

  var indexes = this.whichIndexes(this.filters);
  if (indexes.length) {
    spec.indexes = {};
    indexes.forEach(function(key) {
      spec.indexes[key] = this.filters[key];
    }, this);
    angular.forEach(this.filters, function(value, key) {
      if (typeof(spec.indexes[key]) === 'undefined') {
        if (!spec.filters) {
          spec.filters = {};
        }
        spec.filters[key] = value;
      }
    });
  } else {
    spec.filters = this.filters;
  }
  return spec;
};

StoreQuery.prototype._filterByIndexes = function(values) {
  var matches;
  angular.forEach(values, function(value, key) {
    if (matches && matches.length === 0) {
      return;
    }
    if (typeof (this.indexes[key][value]) === 'undefined') {
      matches = [];
      return;
    }
    var objects = this.indexes[key][value];
    if (!matches) {
      matches = objects.slice(0);
      return;
    }
    matches = this._intersection(matches, objects);
  }, this);
  return matches;
};

StoreQuery.prototype._intersection = function(a, b) {
  var args = arguments;
  return a.filter(function(obj) {
    for (var i = 1; i < args.length; i++) {
      if (args[i].indexOf(obj) === -1) {
        return false;
      }
    }
    return true;
  });
};

StoreQuery.prototype.whichIndexes = function() {
  if (!this.filters || typeof(this.filters) === 'function') {
    return [];
  }
  var indexes = [];
  Object.keys(this.indexes).forEach(function(key) {
    if (typeof(this.filters[key]) !== 'undefined' && typeof(this.filters[key]) !== 'function') {
      indexes.push(key);
    }
  }, this);
  return indexes;
};

StoreQuery.mergeFilters = function(a, b) {
  if (typeof(a) === 'function' && typeof(b) === 'function') {
    return function(obj) {
      return a(obj) && b(obj);
    };
  } else if (!a) {
    return b;
  } else if (!b) {
    return a;
  } else if (typeof(a) === 'function' || typeof(b) === 'function') {
    return function(obj) {
      return StoreQuery.matchObject(obj, a) && StoreQuery.matchObject(obj, b);
    };
  }
  return angular.extend({}, a, b);
};

StoreQuery.matchValue = function(value, cmp, meta) {
  if (typeof(cmp) === 'function') {
    return cmp(value, meta);
  }
  return value === cmp;
};

StoreQuery.matchObjectProperty = function(obj, prop, cmp, meta) {
  if (typeof(obj[prop]) === 'undefined') {
    if (typeof(cmp) === 'function') {
      return cmp(undefined, meta);
    }
    return false;
  }
  return StoreQuery.matchValue(obj[prop], cmp, meta);
};

StoreQuery.matchObject = function(obj, props, meta) {
  if (meta === undefined) {
    meta = {};
  }
  if (typeof(props) === 'function') {
    return props(obj, meta);
  }
  return Object.keys(props).every(key => StoreQuery.matchObjectProperty(obj, key, props[key], meta));
};

StoreQuery.matchArray = function(arr, cmp) {
  if (!arr) { return false; }
  for (var i = 0; i < arr.length; i++) {
    if (cmp(arr[i])) {
      return true;
    }
  }
  return false;
};

StoreQuery.matchStringFuzzy = function(value) {
  return function(propValue) {
    return propValue == value || propValue.toLowerCase() === value.toLowerCase()
      || propValue.replace(/^\s*/, '').replace(/\s*$/, '').toLowerCase() === value.replace(/^\s*/, '').replace(/\s*$/, '').toLowerCase()
      || propValue.replace(/ /g, '').toLowerCase() === value.replace(/ /g, '').toLowerCase();
  };
};


StoreQuery.optional = function(cmp) {
  return function(propValue, meta) {
    return propValue === undefined || cmp === undefined || StoreQuery.matchValue(propValue, cmp, meta);
  };
};

StoreQuery.neq = function(value) {
  return function(propValue) {
    return propValue !== value;
  };
};

StoreQuery.isDefined = function() {
  return function(propValue) {
    return propValue !== undefined;
  };
};

StoreQuery.isUndefined = function() {
  return function(propValue) {
    return propValue === undefined;
  };
};

StoreQuery.isNull = function() {
  return function(propValue) {
    return propValue === undefined || propValue === null;
  };
};

StoreQuery.notNull = function() {
  return function(propValue) {
    return propValue !== undefined && propValue !== null;
  };
};

StoreQuery.isTruthy = function() {
  return function(propValue) {
    return !!propValue;
  };
};

StoreQuery.isFalsy = function() {
  return function(propValue) {
    return !propValue;
  };
};

StoreQuery.contains = function(value) {
  return function(propValue) {
    return propValue !== undefined && propValue.indexOf(value) !== -1;
  };
};

StoreQuery.any = function(values) {
  return function(propValue) {
    return values.indexOf(propValue) !== -1;
  };
};

StoreQuery.arrayMatch = function(cmp) {
  return function(propValue, meta) {
    return StoreQuery.matchArray(propValue, function(value) {
      return StoreQuery.matchValue(value, cmp, meta);
    });
  };
}

StoreQuery.objMatch = function(props) {
  return function(propValue, meta) {
    return propValue !== undefined && propValue !== null && StoreQuery.matchObject(propValue, props, meta);
  };
};

StoreQuery.arrayOfObjMatch = function(props) {
  return function(propValue, meta) {
    return StoreQuery.matchArray(propValue, function(obj) {
      return StoreQuery.matchObject(obj, props, meta);
    });
  };
};

StoreQuery.not = function(cmp) {
  return function(propValue, meta) {
    return !StoreQuery.matchValue(propValue, cmp, meta);
  };
};

StoreQuery.and = function() {
  var args = arguments;
  return function(propValue, meta) {
    for (var i = 0; i < args.length; i++) {
      if (!StoreQuery.matchValue(propValue, args[i], meta)) {
        return false;
      }
    }
    return true;
  };
};

StoreQuery.or = function() {
  var args = arguments;
  return function(propValue, meta) {
    for (var i = 0; i < args.length; i++) {
      if (StoreQuery.matchValue(propValue, args[i], meta)) {
        return true;
      }
    }
    return false;
  };
};
