'use strict';

var StoreResults = function(objects, meta, query) {
  this.meta = meta || {};
  this.query = query;
  if (objects) {
    Array.prototype.splice.apply(this, Array.prototype.concat.apply([0, 0], objects));
  }
};

StoreResults.prototype = new Array();

StoreResults.prototype.toArray = function() {
  return this.slice(0);
};

StoreResults.prototype.sort = function(func, thisArg) {
  var sorted = new StoreResults(this, this.meta);
  Array.prototype.sort.call(sorted, func.bind(thisArg || sorted));
  return sorted;
};

StoreResults.prototype.sortBy = function(props, castToInt) {
  if (!Array.isArray(props)) {
    props = [props];
  }

  props = props.map(prop => {
    let direction = '+';
    if (['+', '-'].includes(prop.charAt(0))) {
      direction = prop.charAt(0);
      prop = prop.substr(1);
    }
    return [prop, direction];
  });

  props.push(['id', '+']);

  return this.sort(function(a, b) {
    for (let i = 0; i < props.length; i++) {
      const [prop, direction] = props[i];
      const valueA = this._getObjectProperty(a, prop);
      const valueB = this._getObjectProperty(b, prop);
      if (valueA === valueB) {
        if (i < props.length - 1) {
          continue;
        }
        return 0;
      }
      const r = this._compare(valueA, valueB, castToInt);
      return direction === '-' ? -r : r;
    }
  });
};

StoreResults.prototype._compare = function(a, b, castToInt) {
  if (a === undefined || a === false) {
    return 1;
  } else if (b === undefined || b === false) {
    return -1;
  } else if (typeof(a) === 'number' && typeof(b) === 'number') {
    return a - b;
  } else if (castToInt) {
    return parseInt(a, 10) - parseInt(b, 10);
  } else if (a && b) {
    return a.localeCompare(b);
  } else {
    return 0;
  }
};

StoreResults.prototype.filter = function(func) {
  return new StoreResults(Array.prototype.filter.apply(this, arguments), this.meta);
};

StoreResults.prototype.filterBy = function(props) {
  return this.filter(function(obj) {
    return objectMatchProps(obj, props);
  });
};

StoreResults.prototype.cleanBy = function(propNames) {
  return this.filter(function(obj) {
    var match = true;
    propNames.forEach(function(prop) {
      if (typeof(obj[prop]) !== 'undefined' || obj[prop] !== null) {
        match = false;
      }
    });
    return match;
  });
};

StoreResults.prototype.keepBy = function(propNames) {
  return this.filter(function(obj) {
    var match = true;
    propNames.forEach(function(prop) {
      if (typeof(obj[prop]) === 'undefined' || obj[prop] === null) {
        match = false;
      }
    });
    return match;
  });
};

StoreResults.prototype.limit = function(max) {
  return new StoreResults(this.slice(0, max), this.meta);
};

StoreResults.prototype.first = function() {
  if (this.length) {
    return this[0];
  }
};

StoreResults.prototype._getObjectProperty = function(obj, prop) {
  if (prop.substr(0, 5) === 'meta:') {
    return this.meta[obj.id][prop.substr(5)];
  }
  var parts = prop.split('.');
  var value = obj;
  for (var i = 0; i < parts.length; i++) {
    if (typeof(value[parts[i]]) === 'undefined') {
      return undefined;
    }
    value = value[parts[i]];
    if (value === null) {
      return null;
    }
  }
  return value;
};

StoreResults.prototype._pluckObject = function(obj, prop) {
  if (typeof(prop) === 'string') {
    return this._getObjectProperty(obj, prop);
  }
  var o = {};
  prop.forEach(function(p) {
    var v = this._getObjectProperty(obj, p);
    if (p.substr(0, 5) === 'meta:') {
      o[p.substr(5)] = v;
    } else {
      o[p] = v;
    }
  }, this);
  return o;
};

StoreResults.prototype.pluck = function(prop) {
  return this.map(function(obj) {
    return this._pluckObject(obj, prop);
  }, this).filter(function(v) {
    return v !== undefined;
  });
};

StoreResults.prototype.pluckFirst = function(prop) {
  if (this.length) {
    return this._pluckObject(this[0], prop);
  }
};

StoreResults.prototype.groupBy = function(prop, pluckProps) {
  var distinct = {};
  this.forEach(function(obj) {
    var key = this._getObjectProperty(obj, prop);
    if (typeof(distinct[key]) === 'undefined') {
      distinct[key] = [];
    }
    if (pluckProps) {
      obj = this._pluckObject(obj, pluckProps);
    }
    distinct[key].push(obj);
  }, this);
  return distinct;
};

StoreResults.prototype.groupByCount = function(prop) {
  var counts = {};
  angular.forEach(this.groupBy(prop), function(objects, key) {
    counts[key] = objects.length;
  });
  return counts;
};

StoreResults.prototype.replace = function(arr) {
  Array.prototype.splice.apply(arr, Array.prototype.concat.apply([0, arr.length], this));
  if (arr instanceof StoreResults) {
    this.meta = arr.meta;
    this.query = arr.query;
  }
};

StoreResults.prototype.refresh = function() {
  if (!this.query) {
    throw new StoreError('Cannot refresh results which were not the results of a query');
  }
  this.query.execute().replace(this);
};

StoreResults.prototype.filterQuery = function(filters) {
  if (!this.query) {
    throw new StoreError('Cannot filters results with query when they were not the results of a query');
  }
  this.query.filter(filters).execute().replace(this);
};

StoreResults.distinct = function(items) {
  var distinct = [];
  for (var i = 0; i < items.length; i++) {
    if (distinct.indexOf(items[i]) === -1) {
      distinct.push(items[i]);
    }
  }
  return distinct;
};
