/* globals Tether */
'use strict';

/*****
* Tether Custom Constraint Module
* ---------------------------------
*
* This module will add a resizable option that helps the element to stay attached to the target
* and continue to be while pinned top or bottom, resizing the element if needed.
*
* This resizable option works only for vertical constraint
* TODO: make it work for horizontal constraint (even if it's not much needed)
*
* A drop-resized CSS class is added for styling purpose.
*
* The change in attachment has been reworked for vertical constraint,
* only triggered when the other attachment option offer more space than the actual to the element.
*
* WARNING: if resizable is enabled, vertical pin and out-of-bound owont be applied
*
* DON'T FORGET: constraints are applied in order.
******/

(function(root, factory) {
  if (typeof define === 'function' && define.amd) {
    define(["tether"], factory);
  } else if (typeof exports === 'object') {
    module.exports = factory(require('tether'));
  } else {
    root.TetherCustomConstraintModule = factory(root.Tether);
  }
}(this, function(Tether) {

var _slicedToArray = (function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i['return']) _i['return'](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError('Invalid attempt to destructure non-iterable instance'); } }; })();

var _TetherBase$Utils = Tether.Utils;
var getBounds = _TetherBase$Utils.getBounds;
var extend = _TetherBase$Utils.extend;
var updateClasses = _TetherBase$Utils.updateClasses;
var removeClass = _TetherBase$Utils.removeClass;
var defer = _TetherBase$Utils.defer;

var BOUNDS_FORMAT = ['left', 'top', 'right', 'bottom'];

function getBoundingRect(tether, to) {
  if (to === 'scrollParent') {
    to = tether.scrollParents[0];
  } else if (to === 'window') {
    to = [pageXOffset, pageYOffset, innerWidth + pageXOffset, innerHeight + pageYOffset];
  }

  if (to === document) {
    to = to.documentElement;
  }

  if (typeof to.nodeType !== 'undefined') {
    (function () {
      var node = to;
      var size = getBounds(to);
      var pos = size;
      var style = getComputedStyle(to);

      to = [pos.left, pos.top, size.width + pos.left, size.height + pos.top];

      // Account any parent Frames scroll offset
      if (node.ownerDocument !== document) {
        var win = node.ownerDocument.defaultView;
        to[0] += win.pageXOffset;
        to[1] += win.pageYOffset;
        to[2] += win.pageXOffset;
        to[3] += win.pageYOffset;
      }

      BOUNDS_FORMAT.forEach(function (side, i) {
        side = side[0].toUpperCase() + side.substr(1);
        if (side === 'Top' || side === 'Left') {
          to[i] += parseFloat(style['border' + side + 'Width']);
        } else {
          to[i] -= parseFloat(style['border' + side + 'Width']);
        }
      });
    })();
  }

  return to;
}

var OFFSET_MAP = {
  top: 0,
  left: 0,
  middle: '50%',
  center: '50%',
  bottom: '100%',
  right: '100%'
};

var attachmentToOffset = function attachmentToOffset(attachment) {
  var left = attachment.left;
  var top = attachment.top;

  if (typeof OFFSET_MAP[attachment.left] !== 'undefined') {
    left = OFFSET_MAP[attachment.left];
  }

  if (typeof OFFSET_MAP[attachment.top] !== 'undefined') {
    top = OFFSET_MAP[attachment.top];
  }

  return { left: left, top: top };
};

function addOffset() {
  var out = { top: 0, left: 0 };

  for (var _len = arguments.length, offsets = Array(_len), _key = 0; _key < _len; _key++) {
    offsets[_key] = arguments[_key];
  }

  offsets.forEach(function (_ref) {
    var top = _ref.top;
    var left = _ref.left;

    if (typeof top === 'string') {
      top = parseFloat(top, 10);
    }
    if (typeof left === 'string') {
      left = parseFloat(left, 10);
    }

    out.top += top;
    out.left += left;
  });

  return out;
}

function offsetToPx(offset, size) {
  if (typeof offset.left === 'string' && offset.left.indexOf('%') !== -1) {
    offset.left = parseFloat(offset.left, 10) / 100 * size.width;
  }
  if (typeof offset.top === 'string' && offset.top.indexOf('%') !== -1) {
    offset.top = parseFloat(offset.top, 10) / 100 * size.height;
  }

  return offset;
}

var TetherCustomConstraintModule = {
  position: function position(_ref) {
    var _this = this;

    var top = _ref.top;
    var left = _ref.left;
    var targetAttachment = _ref.targetAttachment;
    var targetPos = _ref.targetPos;
    var targetOffset = _ref.targetOffset;

    var targetSize = targetPos;
    var targetHeight = targetSize.height;
    var targetWidth = targetSize.width;

    if (!this.options.constraints) {
      return true;
    }

    /*****
    * Custom Constraint management
    * ---------------------------------
    *
    * In order to always get the true height of our element (whether or not
    * it has already been resized to fit in the constraint), we create a mirror dom element
    * that we update each time the position() method is called
    *
    * we also reset the resizedHeight.
    ******/

    var height, width;

    if (this.options.resizable) {
      if (!this.mirror) {
        this.mirror = document.createElement('div');
        this.mirror.className = this.getClass('mirror');
        this.mirror.style.cssText = "position:absolute; top:-999px; left:0; right:auto; bottom:auto; border:0; padding: 0; -moz-box-sizing:content-box; -webkit-box-sizing:content-box; box-sizing:content-box; word-wrap:break-word; height:0 !important; min-height:0 !important; overflow:hidden; transition:none; -webkit-transition:none; -moz-transition:none;";
        document.body.appendChild(this.mirror);
      }

      var clone = this.element.cloneNode(true);
      clone.style.cssText = null;
      // remove the resized class
      var resizedClass;
      if (typeof this.options.resizedClass !== 'undefined') {
        resizedClass = this.options.resizedClass;
      } else {
        resizedClass = this.getClass('resized');
      }

      removeClass(clone, resizedClass);
      this.mirror.innerHTML = '';
      this.mirror.appendChild(clone);

      var mirrorBounds = getBounds(this.mirror.firstChild);
      height = mirrorBounds.height;
      width = mirrorBounds.width;


      // We need to recalculate top and left based on the mirror height and width
      var offset = offsetToPx(attachmentToOffset(this.attachment), { width: width, height: height });
      var manualOffset = offsetToPx(this.offset, { width: width, height: height });
      offset = addOffset(offset, manualOffset);
      left = targetPos.left + targetOffset.left - offset.left;
      top = targetPos.top + targetOffset.top - offset.top;
    } else {
      var _cache = this.cache('element-bounds', function () {
        return getBounds(_this.element);
      });

      height = _cache.height;
      width = _cache.width;
    }

    if (this.options.resizable) {
      this.resizedHeight = null;
    }

    /*****
    * End of custom constraint management
    * -----------------------------------
    ******/

    if (width === 0 && height === 0 && typeof this.lastSize !== 'undefined') {
      var _lastSize = this.lastSize;
      // Handle the item getting hidden as a result of our positioning without glitching
      // the classes in and out
      width = _lastSize.width;
      height = _lastSize.height;
    }

    /*****
    * Custom constraint management
    * ---------------------------------
    *
    * For each constraint you can add a custom resized class to the element
    * and target for CSS purpose.
    *
    ******/

    var allClasses = [this.getClass('pinned'), this.getClass('out-of-bounds'), this.getClass('resized')];

    this.options.constraints.forEach(function (constraint) {
      var outOfBoundsClass = constraint.outOfBoundsClass;
      var pinnedClass = constraint.pinnedClass;
      var resizedClass = constraint.resizedClass;

      if (outOfBoundsClass) {
        allClasses.push(outOfBoundsClass);
      }
      if (pinnedClass) {
        allClasses.push(pinnedClass);
      }

      if (resizedClass) {
        allClasses.push(resizedClass);
      }
    });

    /*****
    * End of custom constraint management
    * -----------------------------------
    ******/

    allClasses.forEach(function (cls) {
      ['left', 'top', 'right', 'bottom'].forEach(function (side) {
        allClasses.push(cls + '-' + side);
      });
    });

    var addClasses = [];

    var tAttachment = extend({}, targetAttachment);
    var eAttachment = extend({}, this.attachment);

    this.options.constraints.forEach(function (constraint, idx) {
      var to = constraint.to;
      var attachment = constraint.attachment;
      var pin = constraint.pin;

      if (typeof attachment === 'undefined') {
        attachment = '';
      }

      var changeAttachX,
          changeAttachY;
      if (attachment.indexOf(' ') >= 0) {
        var _attachment$split = attachment.split(' ');

        var _attachment$split2 = _slicedToArray(_attachment$split, 2);

        changeAttachY = _attachment$split2[0];
        changeAttachX = _attachment$split2[1];
      } else {
        changeAttachX = changeAttachY = attachment;
      }

      var bounds = getBoundingRect(_this, to);

      if (changeAttachY === 'target' || changeAttachY === 'both') {
        if (top < bounds[1] && tAttachment.top === 'top') {
          top += targetHeight;
          tAttachment.top = 'bottom';
        }

        if (top + height > bounds[3] && tAttachment.top === 'bottom') {
          top -= targetHeight;
          tAttachment.top = 'top';
        }
      }

      /*****
      * Custom constraint management
      * ---------------------------------
      *
      * it will change the {top, bottom} attachement
      * only when the change gives more space to the element.
      *
      * It differs from the original constraint module which do the change
      * each time the element leave the "to" (window, scrollParent etc).
      *
      ******/
      if (changeAttachY === 'together') {
        if (tAttachment.top === 'top') {
          if (eAttachment.top === 'bottom' && top < bounds[1] &&
            bounds[3] - (targetSize.top + targetHeight) >= targetSize.top - bounds[1]) {
            top += targetHeight;
            tAttachment.top = 'bottom';
            top += height;
            eAttachment.top = 'top';
          } else if (eAttachment.top === 'top' && top + height > bounds[3] && top - (height - targetHeight) >= bounds[1] &&
            bounds[3] - targetSize.top < (targetSize.top + targetHeight) - bounds[1]) {
            top -= height - targetHeight;
            tAttachment.top = 'bottom';
            eAttachment.top = 'bottom';
          }
        }

        if (tAttachment.top === 'bottom') {
          if (eAttachment.top === 'top' && top + height > bounds[3] &&
            bounds[3] - (targetSize.top + targetHeight) < targetSize.top - bounds[1]) {
            top -= targetHeight;
            tAttachment.top = 'top';
            top -= height;
            eAttachment.top = 'bottom';
          } else if (eAttachment.top === 'bottom' && top < bounds[1] && top + (height * 2 - targetHeight) <= bounds[3] &&
            bounds[3] - targetSize.top >= (targetSize.top + targetHeight) - bounds[1]) {
            top += height - targetHeight;
            tAttachment.top = 'top';
            eAttachment.top = 'top';
          }
        }

        if (tAttachment.top === 'middle') {
          if (top + height > bounds[3] && eAttachment.top === 'top' &&
            bounds[3] - (targetSize.top + targetHeight) < targetSize.top - bounds[1]) {
            top -= height;
            eAttachment.top = 'bottom';
          } else if (top < bounds[1] && eAttachment.top === 'bottom' &&
            bounds[3] - (targetSize.top + targetHeight) >= targetSize.top - bounds[1]) {
            top += height;
            eAttachment.top = 'top';
          }
        }
      }

      if (changeAttachY === 'element' || changeAttachY === 'both') {
        if (top < bounds[1] && eAttachment.top === 'bottom' &&
          (bounds[3] + bounds[1]) / 2 >= height + top) {
          top += height;
          eAttachment.top = 'top';
        }

        if (top + height > bounds[3] && eAttachment.top === 'top' &&
          top > (bounds[3] + bounds[1]) / 2) {
          top -= height;
          eAttachment.top = 'bottom';
        }
      }

      /*****
      * End of custom constraint management
      * -----------------------------------
      ******/

      if (changeAttachX === 'target' || changeAttachX === 'both') {
        if (left < bounds[0] && tAttachment.left === 'left') {
          left += targetWidth;
          tAttachment.left = 'right';
        }

        if (left + width > bounds[2] && tAttachment.left === 'right') {
          left -= targetWidth;
          tAttachment.left = 'left';
        }
      }

      if (changeAttachX === 'together') {
        if (left < bounds[0] && tAttachment.left === 'left') {
          if (eAttachment.left === 'right') {
            left += targetWidth;
            tAttachment.left = 'right';

            left += width;
            eAttachment.left = 'left';
          } else if (eAttachment.left === 'left') {
            left += targetWidth;
            tAttachment.left = 'right';

            left -= width;
            eAttachment.left = 'right';
          }
        } else if (left + width > bounds[2] && tAttachment.left === 'right') {
          if (eAttachment.left === 'left') {
            left -= targetWidth;
            tAttachment.left = 'left';

            left -= width;
            eAttachment.left = 'right';
          } else if (eAttachment.left === 'right') {
            left -= targetWidth;
            tAttachment.left = 'left';

            left += width;
            eAttachment.left = 'left';
          }
        } else if (tAttachment.left === 'center') {
          if (left + width > bounds[2] && eAttachment.left === 'left') {
            left -= width;
            eAttachment.left = 'right';
          } else if (left < bounds[0] && eAttachment.left === 'right') {
            left += width;
            eAttachment.left = 'left';
          }
        }
      }

      if (changeAttachX === 'element' || changeAttachX === 'both') {
        if (left < bounds[0]) {
          if (eAttachment.left === 'right') {
            left += width;
            eAttachment.left = 'left';
          } else if (eAttachment.left === 'center') {
            left += width / 2;
            eAttachment.left = 'left';
          }
        }

        if (left + width > bounds[2]) {
          if (eAttachment.left === 'left') {
            left -= width;
            eAttachment.left = 'right';
          } else if (eAttachment.left === 'center') {
            left -= width / 2;
            eAttachment.left = 'right';
          }
        }
      }

      if (typeof pin === 'string') {
        pin = pin.split(',').map(function (p) {
          return p.trim();
        });
      } else if (pin === true) {
        pin = ['top', 'left', 'right', 'bottom'];
      }

      pin = pin || [];

      var pinned = [];
      var oob = [];

      /*****
      * Custom constraint management
      * ---------------------------------
      *
      * Expose resizedHeight if the options resizable is on.
      * The DOM change of the height has to be done elsewhere
      *
      * Add a custom class to the element and target for CSS purpose
      *
      * WARNING: if resizable is enabled, vertical pin and out-of-bound wont be applied
      *
      * DON'T FORGET: constraints are applied in order.
      ******/
      var resized = [];
      var fixedSidePos;

      // coordinates of the fixed side for further calculation
      // (depending of the element attachment side)
      if (eAttachment.top === 'top') {
        fixedSidePos = top;
      } else if (eAttachment.top === 'bottom') {
        fixedSidePos = top + height;
      } else {
        fixedSidePos = top + (height / 2);
      }

      if (_this.options.resizable) {
        if (top < bounds[1]) {
          top = bounds[1];
          if (eAttachment.top === 'bottom') {
            _this.resizedHeight = fixedSidePos - top;
          }
          resized.push('top');
        }

        if (top + height > bounds[3]) {
          if (eAttachment.top === 'top') {
            top = Math.max(fixedSidePos, bounds[1]);
            _this.resizedHeight = Math.min(top + height, bounds[3] - top);
          } else if (eAttachment.top === 'bottom') {
            top = Math.max(bounds[1], Math.min(fixedSidePos, bounds[3]) - height);
            _this.resizedHeight = Math.min(fixedSidePos, bounds[3]) - top;
          }
          resized.push('bottom');
        }

        // For the next constraints
        if (_this.resizedHeight !== null) {
          height = _this.resizedHeight;
        }
      }


      if (resized.length) {
        (function () {
          var resizedClass;
          if (typeof _this.options.resizedClass !== 'undefined') {
            resizedClass = _this.options.resizedClass;
          } else {
            resizedClass = _this.getClass('resized');
          }

          addClasses.push(resizedClass);
          resized.forEach(function (side) {
            addClasses.push(resizedClass + '-' + side);
          });
        })();
      }

      if (!_this.options.resizable) {
        if (top < bounds[1]) {
          if (pin.indexOf('top') >= 0) {
            top = bounds[1];
            pinned.push('top');
          } else {
            oob.push('top');
          }
        }

        if (top + height > bounds[3]) {
          if (pin.indexOf('bottom') >= 0) {
            top = bounds[3] - height;
            pinned.push('bottom');
          } else {
            oob.push('bottom');
          }
        }
      }

      /*****
      * End of custom constraint management
      * -----------------------------------
      ******/


      if (left < bounds[0]) {
        if (pin.indexOf('left') >= 0) {
          left = bounds[0];
          pinned.push('left');
        } else {
          oob.push('left');
        }
      }

      if (left + width > bounds[2]) {
        if (pin.indexOf('right') >= 0) {
          left = bounds[2] - width;
          pinned.push('right');
        } else {
          oob.push('right');
        }
      }

      if (pinned.length) {
        (function () {
          var pinnedClass;
          if (typeof _this.options.pinnedClass !== 'undefined') {
            pinnedClass = _this.options.pinnedClass;
          } else {
            pinnedClass = _this.getClass('pinned');
          }

          addClasses.push(pinnedClass);
          pinned.forEach(function (side) {
            addClasses.push(pinnedClass + '-' + side);
          });
        })();
      }

      if (oob.length) {
        (function () {
          var oobClass;
          if (typeof _this.options.outOfBoundsClass !== 'undefined') {
            oobClass = _this.options.outOfBoundsClass;
          } else {
            oobClass = _this.getClass('out-of-bounds');
          }

          addClasses.push(oobClass);
          oob.forEach(function (side) {
            addClasses.push(oobClass + '-' + side);
          });
        })();
      }

      if (pinned.indexOf('left') >= 0 || pinned.indexOf('right') >= 0) {
        eAttachment.left = tAttachment.left = false;
      }

      if (pinned.indexOf('top') >= 0 || pinned.indexOf('bottom') >= 0) {
        eAttachment.top = tAttachment.top = false;
      }

      if (tAttachment.top !== targetAttachment.top || tAttachment.left !== targetAttachment.left || eAttachment.top !== _this.attachment.top || eAttachment.left !== _this.attachment.left) {
        _this.updateAttachClasses(eAttachment, tAttachment);
        _this.trigger('update', {
          attachment: eAttachment,
          targetAttachment: tAttachment
        });
      }
    });

    defer(function () {
      if (_this.options.addTargetClasses !== false) {
        updateClasses(_this.target, addClasses, allClasses);
      }
      updateClasses(_this.element, addClasses, allClasses);
    });

    return { top: top, left: left };
  }
};
return TetherCustomConstraintModule;

}));